@@ -32,6 +32,12 @@ def cast_tuple(val, size = 1):
 def calc_same_padding(kernel_size, dilation = 1):
     return dilation * (kernel_size - 1) // 2
 
+def padding_to_multiple_of(n, mult):
+    remainder = n % mult
+    if remainder == 0:
+        return 0
+    return mult - remainder
+
 # decorators
 
 def eval_decorator(fn):
@@ -149,6 +155,34 @@ def apply_rotary_pos_emb(freqs, t):
 
 # helper classes
 
+class ShiftAudioTokens(nn.Module):
+    def __init__(
+        self,
+        fn,
+        audio_tokens_per_timestep = 1
+    ):
+        super().__init__()
+        self.fn = fn
+        self.audio_tokens_per_timestep = audio_tokens_per_timestep
+
+    def forward(self, x, **kwargs):
+        n = x.shape[1]
+
+        # pad to nearest time step
+
+        padding = self.audio_tokens_per_timestep - (n % self.audio_tokens_per_timestep)
+        x = F.pad(x, (0, 0, 0, padding), value = 0.)
+
+        # shift along time
+
+        x_shift, x = x.chunk(2, dim = -1)
+        x_shift = F.pad(x_shift, (0, 0, 1, -1), value = 0.)
+        x = torch.cat((x_shift, x), dim = -1)
+
+        # remove padding if needed
+
+        return self.fn(x[:, :n], **kwargs)
+
 class ShiftVideoTokens(nn.Module):
     def __init__(
         self,
@@ -428,7 +462,7 @@ class Sparse3DNA(nn.Module):
         bos_only = n == 1
         tokens_per_frame = fmap_size ** 2
 
-        padding = 0 if bos_only else (tokens_per_frame - (n - 1) % tokens_per_frame)
+        padding = padding_to_multiple_of(n - 1, tokens_per_frame)
         num_frames = (n + padding) // tokens_per_frame
 
         # pad for last token in video
@@ -558,6 +592,119 @@ class Sparse3DNA(nn.Module):
         out = rearrange(out, '(b h) n d -> b n (h d)', h = h)
         return self.to_out(out)
 
+class SparseCausal2DNA(nn.Module):
+    def __init__(
+        self,
+        *,
+        dim,
+        height = 1,
+        heads = 8,
+        dim_head = 64,
+        dropout = 0.,
+        kernel_size = 5,
+        dilation = 1
+    ):
+        super().__init__()
+        inner_dim = heads * dim_head
+        self.heads = heads
+        self.scale = dim_head ** -0.5
+
+        self.talking_heads = nn.Conv3d(heads, heads, 1, bias = False)
+        self.dropout = nn.Dropout(dropout)
+        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
+        self.to_out = nn.Linear(inner_dim, dim)
+
+        # handle variables for unfold
+
+        self.height = height # width is the sequence length, time-axis - (batch, seq) -> (batch, time, height)
+
+        self.kernel_size = (kernel_size, height)
+        self.dilation = dilation
+        self.padding = calc_same_padding(kernel_size, dilation)
+
+    def forward(
+        self,
+        x,
+        **kwargs
+    ):
+        b, n, h, device = x.shape[0], x.shape[1], self.heads, x.device
+
+        kernel_size, dilation, padding = self.kernel_size, self.dilation, self.padding
+
+        tokens_per_timestep = self.height
+        kernel_numel = kernel_size[0] * kernel_size[1]
+
+        # pad to right length
+
+        bos_only = n == 1
+        seq_pad = padding_to_multiple_of(n - 1, tokens_per_timestep)
+
+        # pad for last token in video
+
+        padded_x = F.pad(x, (0, 0, 0, seq_pad), value = 0.) if seq_pad > 0 else x
+
+        # mask value
+
+        mask_value = -torch.finfo(x.dtype).max
+
+        # derive queries, keys, values
+
+        q, k, v = self.to_qkv(padded_x).chunk(3, dim = -1)
+
+        # handle bos only
+
+        if bos_only:
+            return self.to_out(v)
+
+        out_bos = v[:, :1]
+
+        # split heads
+
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
+
+        # scale
+
+        q = q * self.scale
+
+        # handle bos
+
+        (q_bos, q), (k_bos, k), (v_bos, v) = map(lambda t: (t[:, :, 0], t[:, :, 1:]), (q, k, v))
+
+        # reshape key / values to be unfolded
+
+        k, v = map(lambda t: rearrange(t, 'b h (x y) d -> (b h) d x y ', y = tokens_per_timestep), (k, v))
+        k, v = map(lambda t: F.unfold(t, kernel_size = kernel_size, dilation = (dilation, 1), padding = (padding, 0)), (k, v))
+        k, v = map(lambda t: rearrange(t, '(b h f) (d j) i -> b h i (f j) d', b = b, h = h, j = kernel_numel), (k, v))
+
+        # add bos
+
+        k_bos_repeated, v_bos_repeated = map(lambda t: repeat(t, 'b h d -> b h i 1 d', i = k.shape[-3]), (k_bos, v_bos))
+        k = torch.cat((k_bos_repeated, k), dim = -2)
+        v = torch.cat((v_bos_repeated, v), dim = -2)
+
+        q = rearrange(q, 'b h (x y) d -> b h x y d', y = tokens_per_timestep)
+
+        sim = einsum('b h n i d, b h n j d -> b h n i j', q, k)
+
+        # todo - complete causal masking
+
+        # attention
+
+        attn = stable_softmax(sim, dim = -1)
+        attn = self.talking_heads(attn)
+        attn = self.dropout(attn)
+
+        # aggregate, merge, and combine heads
+
+        out = einsum('b h n i j, b h n j d -> b h n i d', attn, v)
+        out = rearrange(out, 'b h x y d -> b (x y) (h d)')
+
+        # add output for bos back
+
+        out = torch.cat((out_bos, out), dim = -2)
+
+        return self.to_out(out[:, :n])
+
 class SparseCross2DNA(nn.Module):
     def __init__(
         self,
@@ -664,8 +811,7 @@ class SparseCross2DNA(nn.Module):
 
         # pad queries to nearest frame
 
-        q_remainder = q.shape[-2] % tokens_per_frame
-        q_padding = 0 if q_remainder == 0 else (tokens_per_frame - q_remainder)
+        q_padding = padding_to_multiple_of(q.shape[-2], tokens_per_frame)
         q = F.pad(q, (0, 0, 0, q_padding), value = 0.)
 
         # similarity
@@ -706,12 +852,6 @@ For efficient audio <-> video attention
 Largely inspired by chunk cross attention from https://arxiv.org/abs/2112.04426
 """
 
-def padding_to_multiple_of(n, mult):
-    remainder = n % mult
-    if remainder == 0:
-        return 0
-    return mult - remainder
-
 class CrossModalityCrossAttention(nn.Module):
     def __init__(
         self,
@@ -1107,48 +1247,38 @@ class DualModalityDecoder(nn.Module):
         depth,
         num_audio_tokens_per_video_frame,
         num_video_tokens_per_frame,
+        sparse_3dna_video_shape,
         heads = 8,
         dim_head = 64,
         ff_mult = 4,
         attn_dropout = 0.,
         ff_dropout = 0.,
         ff_chunk_size = None,
-        sparse_3dna_attn = False,
         sparse_3dna_kernel_size = 3,
-        sparse_3dna_video_shape = None,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
         shift_video_tokens = False,
+        shift_audio_tokens = False,
+        audio_tokens_per_timestep = 1,
         cross_modality_attn_every = 3
     ):
         super().__init__()
-        assert not (sparse_3dna_attn and not exists(sparse_3dna_video_shape)), 'sparse_3dna_video_shape must be defined if turned on'
-
         self.layers = MList([])
         self.layer_types = []
 
-        def intra_modality_attn(sparse_3dna_attn):
-            if sparse_3dna_attn:
-                dilation = sparse_3dna_dilations[ind % len(sparse_3dna_dilations)]
+        def intra_modality_attn(is_video):
+            dilation = sparse_3dna_dilations[ind % len(sparse_3dna_dilations)]
 
-                self_attn = Sparse3DNA(
-                    dim = dim,
-                    heads = heads,
-                    dim_head = dim_head,
-                    causal = True,
-                    kernel_size = sparse_3dna_kernel_size,
-                    dilation = dilation,
-                    video_shape = sparse_3dna_video_shape,
-                    query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
-                )
-            else:
-                self_attn = Attention(
-                    dim = dim,
-                    heads = heads,
-                    dim_head = dim_head,
-                    causal = True,
-                    dropout = attn_dropout
-                )
+            self_attn = Sparse3DNA(
+                dim = dim,
+                heads = heads,
+                dim_head = dim_head,
+                causal = True,
+                kernel_size = sparse_3dna_kernel_size,
+                dilation = dilation,
+                video_shape = sparse_3dna_video_shape,
+                query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+            )
 
             cross_attn = Attention(
                 dim = dim,
@@ -1159,11 +1289,15 @@ class DualModalityDecoder(nn.Module):
 
             ff = FeedForward(dim = dim, mult = ff_mult, dropout = ff_dropout, chunk_size = ff_chunk_size)
 
-            if sparse_3dna_attn and shift_video_tokens:
+            if is_video and shift_video_tokens:
                 fmap_size = sparse_3dna_video_shape[-1]
                 self_attn = ShiftVideoTokens(self_attn, image_size = fmap_size)
                 ff        = ShiftVideoTokens(ff, image_size = fmap_size)
 
+            if not is_video and shift_audio_tokens:
+                self_attn = ShiftAudioTokens(self_attn, audio_tokens_per_timestep = audio_tokens_per_timestep)
+                ff        = ShiftAudioTokens(ff, audio_tokens_per_timestep = audio_tokens_per_timestep)
+
             return MList([
                 SandwichNorm(dim = dim, fn = self_attn),
                 SandwichNorm(dim = dim, fn = cross_attn),
@@ -1189,8 +1323,8 @@ class DualModalityDecoder(nn.Module):
             ])
 
         for ind in range(depth):
-            video_modality_attn = intra_modality_attn(sparse_3dna_attn = sparse_3dna_attn)
-            audio_modality_attn = intra_modality_attn(sparse_3dna_attn = False)
+            video_modality_attn = intra_modality_attn(True)
+            audio_modality_attn = intra_modality_attn(False)
 
             self.layer_types.append('intra_modality')
 
@@ -1270,23 +1404,22 @@ class ReversibleDualModalityDecoder(nn.Module):
         depth,
         num_audio_tokens_per_video_frame,
         num_video_tokens_per_frame,
+        sparse_3dna_video_shape,
         heads = 8,
         dim_head = 64,
         ff_mult = 4,
         attn_dropout = 0.,
         ff_dropout = 0.,
         ff_chunk_size = None,
-        sparse_3dna_attn = False,
         sparse_3dna_kernel_size = 3,
-        sparse_3dna_video_shape = None,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
         shift_video_tokens = False,
+        shift_audio_tokens = False,
+        audio_tokens_per_timestep = 1,
         cross_modality_attn_every = 3
     ):
         super().__init__()
-        assert not (sparse_3dna_attn and not exists(sparse_3dna_video_shape)), 'sparse_3dna_video_shape must be defined if turned on'
-
         self.layers = MList([])
         self.layer_types = []
 
@@ -1318,11 +1451,15 @@ class ReversibleDualModalityDecoder(nn.Module):
             video_ff = create_ff()
             audio_ff = create_ff()
 
-            if sparse_3dna_attn and shift_video_tokens:
+            if shift_video_tokens:
                 fmap_size = sparse_3dna_video_shape[-1]
                 video_self_attn = ShiftVideoTokens(video_self_attn, image_size = fmap_size)
                 video_ff        = ShiftVideoTokens(video_ff, image_size = fmap_size)
 
+            if shift_audio_tokens:
+                audio_self_attn = ShiftAudioTokens(audio_self_attn, audio_tokens_per_timestep = audio_tokens_per_timestep)
+                audio_ff        = ShiftAudioTokens(audio_ff, audio_tokens_per_timestep = audio_tokens_per_timestep)
+
             self.layers.append(MList([
                 norm_wrapper(video_self_attn),
                 norm_wrapper(video_ff),
@@ -1723,6 +1860,7 @@ class NUWAVideoAudio(nn.Module):
         image_size,
         num_audio_tokens,
         num_audio_tokens_per_video_frame,
+        audio_tokens_per_timestep = 1,
         max_video_frames = 5,
         text_num_tokens = 49408,
         text_max_seq_len = 256,
@@ -1740,6 +1878,7 @@ class NUWAVideoAudio(nn.Module):
         ff_chunk_size = None,
         embed_gradient_frac = 0.2,
         shift_video_tokens = True,
+        shift_audio_tokens = True,
         sparse_3dna_kernel_size = 3,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilation = 1,
@@ -1810,9 +1949,10 @@ class NUWAVideoAudio(nn.Module):
             attn_dropout = attn_dropout,
             ff_dropout = ff_dropout,
             ff_chunk_size = ff_chunk_size,
+            audio_tokens_per_timestep = audio_tokens_per_timestep,
+            shift_audio_tokens = shift_audio_tokens,
             shift_video_tokens = shift_video_tokens,
             sparse_3dna_video_shape = video_shape,
-            sparse_3dna_attn = True,
             sparse_3dna_kernel_size = sparse_3dna_kernel_size,
             sparse_3dna_dilations = sparse_3dna_dilations,
             sparse_3dna_query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
